﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel.Core;
using Windows.UI.Core;
using Windows.UI.Popups;
using DAQCdashboard.Annotations;
using PiPlates;

namespace DAQCdashboard.ViewModels
{
    public class DAQCViewModel: INotifyPropertyChanged
    {
        private DAQCplate _plate;
        private bool _pollingEnabled;
        private Timer _pollingTimer;
        private bool _inputInteruptsEnabled;
        private bool _switchValue;

        private int PollTime = 1000;    //poll every 1000ms
        private int _selectedBoardNumber;
        private float _selectedBoardHWrev;
        private float _selectedBoardFWrev;
        private string _selectedBoardId;
        private short _dOutValue;

        private DAQCplate Plate
        {
            get { return _plate; }
            set
            {
                //clean up previous plate
                if (_plate != null)
                {
                    //disable interupts
                    _plate.disableSWint();
                    _plate.intDisable();

                    //remove interupt event handers
                    _plate.DigitalInputInteruptOccurred -= DigitalInputInteruptOccurred;
                    _plate.SwitchInteruptOccurred -= SwitchInteruptOccurred;

                    _plate.Dispose();
                }

                //save the new value
                _plate = value;

                //get the current values and set up event handlers for the new plate
                if (_plate != null)
                {
                    var inVals = _plate.getDINall();
                    for (int i = 0; i < 8; i++)
                    {
                        var val = (inVals & (1 << i)) != 0;
                        (DigitalInputs.FirstOrDefault(di=>di.Channel == i)).Value = val;                        
                    }

                    var outVals = _plate.getDOUTall();
                    for (int i = 0; i < 7; i++)
                    {
                        var val = (outVals & (1 << i)) != 0;
                        (DigitalOutputs.FirstOrDefault(o => o.Channel == i)).Value = val;
                    }

                    var ainVals = _plate.getADCall();
                    for (int i = 0; i < 8; i++)
                    {
                        AnalogInputs.FirstOrDefault(ai=> ai.Channel == (byte)i).Value = ainVals[i];
                    }
                    
                    for (int i = 0; i < 2; i++)
                    {
                        AnalogOutputs.FirstOrDefault(ao => ao.Channel == (byte)i).Value = 0F;
                    }

                    GreenLed.Value = _plate.getLED(GreenLed.Channel);
                    RedLed.Value = _plate.getLED(RedLed.Channel);

                    SelectedBoardFWrev = _plate.getFWrev();
                    SelectedBoardHWrev = _plate.getHWrev();
                    SelectedBoardId = _plate.getID();
                    _plate.DigitalInputInteruptOccurred += DigitalInputInteruptOccurred;
                    _plate.SwitchInteruptOccurred += SwitchInteruptOccurred;

                    //enable all interupts
                    _plate.intEnable();

                    //enable switch interupt
                    _plate.enableSWint();
                }
            }
        }
        
        #region Public Properties
        public int SelectedBoardNumber
        {
            get { return _selectedBoardNumber; }
            set
            {
                if (value == _selectedBoardNumber) return;
                _selectedBoardNumber = value;
                OnPropertyChanged();

                Plate = new DAQCplate((byte)_selectedBoardNumber);
            }
        }
        public List<int> AvailableBoards { get; set; }

        public float SelectedBoardHWrev
        {
            get { return _selectedBoardHWrev; }
            set
            {
                if (value.Equals(_selectedBoardHWrev)) return;
                _selectedBoardHWrev = value;
                OnPropertyChanged();
            }
        }

        public float SelectedBoardFWrev
        {
            get { return _selectedBoardFWrev; }
            set
            {
                if (value.Equals(_selectedBoardFWrev)) return;
                _selectedBoardFWrev = value;
                OnPropertyChanged();
            }
        }

        public string SelectedBoardId
        {
            get { return _selectedBoardId; }
            set
            {
                if (value == _selectedBoardId) return;
                _selectedBoardId = value;
                OnPropertyChanged();
            }
        }

        public ObservableCollection<DigitalIo> DigitalInputs { get; set; }
        public ObservableCollection<DigitalIo> DigitalOutputs { get; set; }
        public ObservableCollection<AnalogIo> AnalogInputs { get; set; }
        public ObservableCollection<AnalogIo> AnalogOutputs { get; set; }
        public ObservableCollection<AnalogIo> PwmOutputs { get; set; }
        public DigitalIo GreenLed { get; set; }
        public DigitalIo RedLed { get; set; }
        public bool InputInteruptsEnabled
        {
            get { return _inputInteruptsEnabled; }
            set
            {
                if (value == _inputInteruptsEnabled) return;
                _inputInteruptsEnabled = value;
                OnPropertyChanged();
                if (_inputInteruptsEnabled)
                {
                    //disable interupts for all input channels
                    foreach (var di in DigitalInputs)
                    {
                        Plate.enableDINint(di.Channel, InteruptEdgeType.Both);
                    }
                }
                else
                {
                    //disable interupts for all input channels
                    foreach (var di in DigitalInputs)
                    {
                        Plate.disableDINint(di.Channel);
                    }
                }
            }
        }
        public bool SwitchValue
        {
            get { return _switchValue; }
            set
            {
                //if (value == _switchValue) return;
                _switchValue = value;
                OnPropertyChanged();
            }
        }
        public bool AnalogPollingEnabled
        {
            get { return _pollingEnabled; }
            set
            {
                if (value == _pollingEnabled) return;
                _pollingEnabled = value;
                OnPropertyChanged();

                if (_pollingEnabled)
                {
                    _pollingTimer = new Timer(UpdateAnalogValues, null, PollTime, PollTime);
                }
                else
                {
                    _pollingTimer.Change(Timeout.Infinite, Timeout.Infinite);
                    _pollingTimer.Dispose();
                }
            }
        }
        #endregion
        
        #region Constructors

        public DAQCViewModel()
        {

            RedLed = new DigitalIo() {Channel = 0};
            RedLed.PropertyChanged += LedPropertyChanged;
            GreenLed = new DigitalIo() {Channel = 1};
            GreenLed.PropertyChanged += LedPropertyChanged;

            DigitalInputs = new ObservableCollection<DigitalIo>
            {
                new DigitalIo() {Channel = 0, Value = false},
                new DigitalIo() {Channel = 1, Value = false},
                new DigitalIo() {Channel = 2, Value = false},
                new DigitalIo() {Channel = 3, Value = false},
                new DigitalIo() {Channel = 4, Value = false},
                new DigitalIo() {Channel = 5, Value = false},
                new DigitalIo() {Channel = 6, Value = false},
                new DigitalIo() {Channel = 7, Value = false},
            };

            DigitalOutputs = new ObservableCollection<DigitalIo>
            {
                new DigitalIo() {Channel = 0, Value = false},
                new DigitalIo() {Channel = 1, Value = false},
                new DigitalIo() {Channel = 2, Value = false},
                new DigitalIo() {Channel = 3, Value = false},
                new DigitalIo() {Channel = 4, Value = false},
                new DigitalIo() {Channel = 5, Value = false},
                new DigitalIo() {Channel = 6, Value = false},
            };
            foreach (var dout in DigitalOutputs) dout.PropertyChanged += WriteDigitalOutput;

            AnalogInputs = new ObservableCollection<AnalogIo>
            {
                new AnalogIo() {Channel = 0, Value = 0F},
                new AnalogIo() {Channel = 1, Value = 0F},
                new AnalogIo() {Channel = 2, Value = 0F},
                new AnalogIo() {Channel = 3, Value = 0F},
                new AnalogIo() {Channel = 4, Value = 0F},
                new AnalogIo() {Channel = 5, Value = 0F},
                new AnalogIo() {Channel = 6, Value = 0F},
                new AnalogIo() {Channel = 7, Value = 0F},
            };

            AnalogOutputs = new ObservableCollection<AnalogIo>
            {
                new AnalogIo() {Channel = 0, Value = 0F, MinValue = 0F, MaxValue = 4.096F},
                new AnalogIo() {Channel = 1, Value = 0F, MinValue = 0F, MaxValue = 4.096F},
            };
            foreach (var ao in AnalogOutputs) ao.PropertyChanged += WriteAnalogOutput;

            PwmOutputs = new ObservableCollection<AnalogIo>
            {
                new AnalogIo() {Channel = 0, Value = 0F, MinValue = 0F, MaxValue = 100F},
                new AnalogIo() {Channel = 1, Value = 0F, MinValue = 0F, MaxValue = 100F},
            };
            foreach (var pwm in PwmOutputs) pwm.PropertyChanged += WritePwmOutput;

            //create relay command to fetch digital input value
            ReadDigitalInput = new RelayCommand<object>(ReadDigitalInputExecute);

            //create relay command to fetch digital analog value
            ReadAnalogInput = new RelayCommand<object>(ReadAnalogInputExecute);

            WriteAllDigitalOutputs = new RelayCommand<object>(WriteAllOutputsExecute);

            //try to get a list of the available DAQCplates
            while (AvailableBoards == null || !AvailableBoards.Any())
            {
                try
                {
                    var plates = DAQCplate.Poll();
                    AvailableBoards = plates == null ? new List<int>() : new List<int>(plates);
                }
                catch (Exception e)
                {
                    Debug.WriteLine(e.Message);
                }
            }


            //try to create an instance of the first available DAQCplate
            try
            {
                _selectedBoardNumber = AvailableBoards.FirstOrDefault();
                Plate = new DAQCplate((byte)(SelectedBoardNumber));   //default to the first board
            }
            catch (Exception e)
            {
                Debug.WriteLine(e.Message);
            }
        }

        private void WriteAllOutputsExecute(object obj)
        {
            try
            {
                Plate.setDOUTall((byte)DOutValue);

                Task.Delay(TimeSpan.FromSeconds(.1)).Wait();
                //update the individual bits
                var outVals = _plate.getDOUTall();
                for (int i = 0; i < 7; i++)
                {
                    var val = (outVals & (1 << i)) != 0;
                    (DigitalOutputs.FirstOrDefault(o => o.Channel == i)).Value = val;
                }

            }
            catch(Exception e)
            {
                Debug.WriteLine(e.Message);
            }
        }

        public short DOutValue
        {
            get { return _dOutValue; }
            set
            {
                if (value == _dOutValue) return;
                _dOutValue = value;
                OnPropertyChanged();
            }
        }

        #endregion
        private void ReadAnalogInputExecute(object o)
        {
            var ain = o as AnalogIo;
            if (ain == null) return;
            try
            {
                ain.Value = Plate.getADC(ain.Channel);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }
        private void WritePwmOutput(object sender, PropertyChangedEventArgs args)
        {
            var pwm = sender as AnalogIo;
            if (pwm == null) return;
            try
            {
                if (args.PropertyName != "Value") return;

                //value set by user is 0-100%.  Convert this to 0-1024
                var val = (byte)((1024 / 100F) * pwm.Value);
                Plate.setPWM(pwm.Channel, val);
            }
            catch (Exception e)
            {
                Debug.WriteLine(e.Message);
            }
        }
        private void WriteAnalogOutput(object sender, PropertyChangedEventArgs args)
        {
            if (args.PropertyName != "Value") return;

            var aout = sender as AnalogIo;
            if (aout == null) return;
            try
            {
                Plate.setDAC(aout.Channel, aout.Value);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }
        private void WriteDigitalOutput(object sender, PropertyChangedEventArgs args)
        {
            if (args.PropertyName != "Value") return;

            var x = sender as DigitalIo;
            if (x == null) return;
            try
            {
                if (x.Value)
                {
                    Plate.setDOUTbit(x.Channel);
                }
                else
                {
                    Plate.clrDOUTbit(x.Channel);
                }

                Task.Delay(TimeSpan.FromSeconds(.1)).Wait();
                DOutValue = Plate.getDOUTall();
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        private void ReadDigitalInputExecute(object o)
        {
            var din = o as DigitalIo;
            if (din == null) return;

            try
            {
                din.Value = Plate.getDINbit(din.Channel);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }
        private async void DigitalInputInteruptOccurred(object sender, DigitalInputInteruptOccurredEventArgs e)
        {
            //update the value of the digital input that triggered the interupt
            await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(
                CoreDispatcherPriority.Normal, () =>
                {
                    var din = DigitalInputs.FirstOrDefault(di => di.Channel == e.Channel);
                    if (din == null) return;
                    din.Value = e.Value;
                });
        }
        private async void SwitchInteruptOccurred(object sender, EventArgs e)
        {
            //toggle switch value ever time the switch is pressed
            await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(
                    CoreDispatcherPriority.Normal, () => SwitchValue = !SwitchValue);
        }
        private void LedPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var led = sender as DigitalIo;
            if (led == null) return;
            try
            {
                if (led.Value)
                {
                    Plate.setLED(led.Channel);
                }
                else
                {
                    Plate.clrLED(led.Channel);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        #region Commands
        public RelayCommand<object> ReadDigitalInput { get; set; }
        public RelayCommand<object> ReadAnalogInput { get; set; }

        public RelayCommand<object> WriteAllDigitalOutputs { get; set; } 
        #endregion

        private async void UpdateAnalogValues(object state)
        {
            //pause the timer so we don't try to update while we are already in an update
            _pollingTimer.Change(Timeout.Infinite, Timeout.Infinite);
            try
            {
                await Windows.ApplicationModel.Core.CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(
                    CoreDispatcherPriority.Normal, () =>
                    {
                        try
                        {
                            //get all analog input values
                            var ains = Plate.getADCall();
                            //update each value
                            for (int i = 0; i < ains.Length; i++)
                            {
                                var ai = AnalogInputs.FirstOrDefault(x => x.Channel == i);
                                if (ai != null)
                                {
                                    ai.Value = ains[i];
                                }
                            }
                        }
                        catch (Exception e)
                        {
                            Debug.WriteLine(e);
                        }
                    });
            }
            finally
            {
                //restart the timer
                _pollingTimer.Change(PollTime, PollTime);
            }
        }

        #region INotifyPropertyChanged

        public event PropertyChangedEventHandler PropertyChanged;

        [NotifyPropertyChangedInvocator]
        protected virtual void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

    }
}
